/*
 nRF24 Multi-Network TX/RX
 Version: 0.3
 Author: Alex from insideGadgets (http://www.insidegadgets.com)
 Created: 25/05/2013
 Last Modified: 29/10/2013
 
 */
 
 /*
	Copyright (c) 2007 Stefan Engelke <mbox@stefanengelke.de>

	Permission is hereby granted, free of charge, to any person 
	obtaining a copy of this software and associated documentation 
	files (the "Software"), to deal in the Software without 
	restriction, including without limitation the rights to use, copy, 
	modify, merge, publish, distribute, sublicense, and/or sell copies 
	of the Software, and to permit persons to whom the Software is 
	furnished to do so, subject to the following conditions:
	
	The above copyright notice and this permission notice shall be 
	included in all copies or substantial portions of the Software.
	
	THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
	EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
	MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
	NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT 
	HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, 
	WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, 
	OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER 
	DEALINGS IN THE SOFTWARE.
	
	$Id$
*/

#include "mirf.h"
#include "nRF24L01.h"
#include <avr/io.h>
#include <avr/interrupt.h>

// Defines for setting the MiRF registers for transmitting or receiving mode
#define TX_POWERUP mirf_config_register(CONFIG, mirf_CONFIG | ( (1<<PWR_UP) | (0<<PRIM_RX) ) )
#define RX_POWERUP mirf_config_register(CONFIG, mirf_CONFIG | ( (1<<PWR_UP) | (1<<PRIM_RX) ) )
#define POWERDOWN mirf_config_register(CONFIG, mirf_CONFIG | ( (0<<PWR_UP) ) )

extern uint8_t data_in[];
extern uint8_t data_out[];
extern uint8_t neighbourNRF[];
extern uint8_t neighbourNRFcount;

// Initialise pins
void mirf_init(void) {
	// Define CSN and CE as Output and set them to default
	DDRB |= ((1<<CSN)|(1<<CE));
	mirf_CE_lo;
	mirf_CSN_hi;
}

// Sets the important registers in the MiRF module
void mirf_config(void) {
	// Set RF channel	
	mirf_config_register(RF_CH, mirf_CH);
	
	// Set length of incoming payload 
	mirf_config_register(RX_PW_P0, mirf_PAYLOAD);
	
	// Disable auto acknowledgement
	mirf_config_register(EN_AA, 0<<ENAA_P0);
	
	// Disable auto retransmit delay
	mirf_config_register(SETUP_RETR, 0);
	
	// Set mirf_ADDR
	mirf_write_register(RX_ADDR_P0, mirf_ADDR, 5);
	mirf_write_register(TX_ADDR, mirf_ADDR, 5);
	
	// Enable RX_ADDR_P0 address matching
	mirf_config_register(EN_RXADDR, (1<<ERX_P0));
	
	// Power up in transmitter mode
	RX_POWERUP;
}

// Flush RX and TX FIFO
void mirf_flush_rx_tx(void) {
	mirf_CSN_lo; // Pull down chip select
	spi_transfer(FLUSH_RX); // Flush RX
	mirf_CSN_hi; // Pull up chip select

	mirf_CSN_lo; // Pull down chip select
	spi_transfer(FLUSH_TX);  // Write cmd to flush tx fifo
	mirf_CSN_hi; // Pull up chip select
}

// Read the status register
uint8_t mirf_status(void) {
	mirf_CSN_lo; // Pull down chip select
	spi_transfer(R_REGISTER | (REGISTER_MASK & STATUS));
	uint8_t status = spi_transfer(NOP); // Read status register
	mirf_CSN_hi; // Pull up chip select
	return status;
}

// Read the FIFO status register
uint8_t mirf_fifostatus(void) {
	mirf_CSN_lo; // Pull down chip select
	spi_transfer(R_REGISTER | (REGISTER_MASK & FIFO_STATUS));
	uint8_t status = spi_transfer(NOP); // Read status register
	mirf_CSN_hi; // Pull up chip select
	return status;
}

// Checks if data is available for reading
uint8_t mirf_data_ready(void) {
	mirf_CSN_lo; // Pull down chip select
	spi_transfer(R_REGISTER | (REGISTER_MASK & STATUS));
	uint8_t status = spi_transfer(NOP); // Read status register
	mirf_CSN_hi; // Pull up chip select
	return status & (1<<RX_DR);
}

// Checks if MAX_RT has been reached
uint8_t mirf_max_rt_reached(void) {
	mirf_CSN_lo; // Pull down chip select
	spi_transfer(R_REGISTER | (REGISTER_MASK & STATUS));
	uint8_t status = spi_transfer(NOP); // Read status register
	mirf_CSN_hi; // Pull up chip select
	return status & (1<<MAX_RT);
}

// Write one byte into the MiRF register
void mirf_config_register(uint8_t reg, uint8_t value) {
	mirf_CSN_lo;
	spi_transfer(W_REGISTER | (REGISTER_MASK & reg));
	spi_transfer(value);
	mirf_CSN_hi;
}

// Reads an array of bytes from the MiRF registers.
void mirf_read_register(uint8_t reg, uint8_t *value, uint8_t len) {
	mirf_CSN_lo;
	spi_transfer(R_REGISTER | (REGISTER_MASK & reg));
	spi_read_data(value, len);
	mirf_CSN_hi;
}

// Writes an array of bytes into the MiRF register
void mirf_write_register(uint8_t reg, uint8_t *value, uint8_t len) {
	mirf_CSN_lo;
	spi_transfer(W_REGISTER | (REGISTER_MASK & reg));
	spi_write_data(value, len);
	mirf_CSN_hi;
}

// Check for neighbouring nRF modules, returns the number of nRF modules found
uint8_t check_neighbouring_nrf(void) {
	neighbourNRFcount = 0;
	
	// Clear data out
	for (int x = 0; x < 32; x++) {
		data_out[x] = 0;
	}
	
	data_out[FROM] = MYADDR;
	data_out[TO] = ALLADDR; // Send to All addresses
	data_out[TYPE] = NEIGHBOURREQ; // Request all nRFs to reply back
	data_out[DATA] = 0;
	
	mirf_transmit_data(NODELAY);
	
	RX_POWERUP; // Power up to receiver mode
	mirf_flush_rx_tx(); // Flush RX and TX FIFO
	mirf_CE_hi; // Start listening
	
	// Wait for replies for 50 ms
	// To do: Add a proper timer
	for (int x = 0; x < 30; x++) {
		byte waitcount = 0;
		while (!(mirf_status() & (1<<RX_DR))) { // Wait until data arrives
			_delay_us(50);
			waitcount++;
			if (waitcount >= 100) { // If we haven't heard back in 5ms
				break;
			}
		}
		
		// Check all FIFO RX
		while (!(mirf_fifostatus() & (1<<RX_EMPTY))) {
			// Clear the data_in
			for (int x = 0; x < 32; x++) {
				data_in[x] = 0;
			}
			
			// Read the payload
			mirf_CSN_lo; // Pull down chip select
			spi_transfer(R_RX_PAYLOAD); // Send cmd to read rx payload
			spi_read_data(data_in, mirf_PAYLOAD); // Read payload
			mirf_CSN_hi; // Pull up chip select
			mirf_config_register(STATUS,(1<<RX_DR)); // Reset status register
			
			if (data_in[TO] == MYADDR && data_in[TYPE] == NEIGHBOUR_RESPOND) { 
				// Check the from address doesn't exist already
				bool fromExists = false;
				for (int x = 0; x < neighbourNRFcount; x++) {
					if (neighbourNRF[x] == data_in[FROM]) {
						fromExists = true;
						break;
					}
				}
				if (fromExists == false) { // Add to list
					neighbourNRF[neighbourNRFcount] = data_in[FROM];
					neighbourNRFcount++;
				}
			}
		}
	}
	mirf_CE_lo; 
	
	// Blink LED for each nRF that replies back
	for (int x = 0; x < neighbourNRFcount; x++) {
		PORTB |= (1<<PB2);
		_delay_ms(500);
		PORTB &= ~(1<<PB2);
		_delay_ms(500);
	}
	
	return neighbourNRFcount;
}

// Read received packet and process it
uint8_t read_packet_and_process(void) {
	
	// Clear the data_in
	for (int x = 0; x < 32; x++) {
		data_in[x] = 0;
	}
	
	// Read the payload
	mirf_CSN_lo; // Pull down chip select
	spi_transfer(R_RX_PAYLOAD); // Send cmd to read rx payload
	spi_read_data(data_in, mirf_PAYLOAD); // Read payload
	mirf_CSN_hi; // Pull up chip select
	mirf_config_register(STATUS,(1<<RX_DR)); // Reset status register
	
	// Addr 1
	if (data_in[TO] == MYADDR && data_in[TYPE] == FWD && data_in[TTL] == 0) {		
		// Clear data out
		for (int x = 0; x < 32; x++) {
			data_out[x] = 0;
		}
		
		data_out[FROM] = MYADDR; // From our address
		data_out[TO] = data_in[FROM]; // Send back to the address that we received this packet from
		data_out[TYPE] = FWDRESPOND; // Ack respond
		data_out[TTL] = 1;
		
		// Send the ACK to the forward with random delay in-between
		mirf_transmit_data(RANDOMDELAY);
		
		return RECEIVED_DATA_FROM_FORWARD;
	}
	
	return 0;
}

// Listen for packets and timeout if we reach 5ms
uint8_t mirf_listen_5ms_timeout(void) {
	RX_POWERUP; // Power up to receiver mode
	mirf_flush_rx_tx(); // Flush RX and TX FIFO
	mirf_CE_hi; // Start listening
	
	byte waitcount = 0;
	while (!(mirf_status() & (1<<RX_DR))) { // Wait until data arrives
		_delay_us(50);
		waitcount++;
		if (waitcount >= 100) { // If we haven't heard back in 5ms, exit
			mirf_CE_lo; // Stop listening
			return 0;
		}
	}
	mirf_CE_lo; // Stop listening
	
	return 1;
}

// Send data as many times as specified and add a random delay if we know that other nRFs will be transmitting
void mirf_transmit_data(bool randomDelay) {
	TX_POWERUP; // Power up to transmitter mode
	mirf_flush_rx_tx(); // Flush RX and TX FIFO
	
	mirf_CSN_lo;  // Pull down chip select
	spi_transfer(W_TX_PAYLOAD); // Write cmd to write payload
	spi_write_data(data_out, mirf_PAYLOAD); // Write payload
	mirf_CSN_hi; // Pull up chip select
	
	// Reuse TX Payload, re-send the packet a few times to ensure they get it
	for (int x = 0; x < 4; x++) {
		mirf_CE_hi; // Start transmission
		_delay_us(15);
		mirf_CE_lo;
		
		_delay_us(500); // Wait until packet is transmitted
		
		// Wait a random amount of time (0 to 3.726ms)
		if (randomDelay == true) {
			_delay_us(rand() / 10);
		}
		
		mirf_CSN_lo;  // Pull down chip select
		spi_transfer(REUSE_TX_PL); // Write cmd to reuse last payload
		mirf_CSN_hi; // Pull up chip select
	}
	
	// We always flush TX/RX before sending/receiving so we don't need to do it here
}

// Send data 4 times if needed (4ms) until we receive an ACK back
uint8_t mirf_transmit_data_check_ack(void) {
	TX_POWERUP; // Power up to transmitter mode
	mirf_flush_rx_tx(); // Flush RX and TX FIFO
	
	for (int x = 0; x < 32; x++) {
		data_in[x] = 0; // Clear the data_in
	}
	
	mirf_CSN_lo;  // Pull down chip select
	spi_transfer(W_TX_PAYLOAD); // Write cmd to write payload
	spi_write_data(data_out, mirf_PAYLOAD); // Write payload
	mirf_CSN_hi; // Pull up chip select
	
	// Reuse TX Payload, re-send the packet a few times to ensure they get it
	for (int x = 0; x < 4; x++) {
		TX_POWERUP; // Power up to transmitter mode
		
		mirf_CE_hi; // Start transmission
		_delay_us(15);
		mirf_CE_lo;
		
		_delay_us(500); // Wait until packet is transmitted
		
		mirf_CSN_lo;  // Pull down chip select
		spi_transfer(REUSE_TX_PL); // Write cmd to reuse last payload
		mirf_CSN_hi; // Pull up chip select
		
		RX_POWERUP;
		mirf_CE_hi; // Start listening
		
		uint8_t waitcount = 0;
		while (!(mirf_status() & (1<<RX_DR))) { // Wait until data arrives
			_delay_us(50);
			waitcount++;
			if (waitcount >= 10) { // If we haven't heard back in ~500uS, exit
				mirf_CE_lo; // Stop listening
				break;
			}
		}
		
		if (mirf_status() & (1<<RX_DR)) { // Check if any packets received
			mirf_CSN_lo; // Pull down chip select
			spi_transfer(R_RX_PAYLOAD); // Send cmd to read rx payload
			spi_read_data(data_in, mirf_PAYLOAD); // Read payload
			mirf_CSN_hi; // Pull up chip select
			mirf_config_register(STATUS,(1<<RX_DR)); // Reset status register
			
			// Check the ACK matches our address, who it came from and the ack type
			if (data_in[TO] == MYADDR && data_in[FROM] == data_out[TO] && data_in[TYPE] == ACKRESPOND) {
				return 1;
			}
		}
	}
	
	// We always flush TX/RX before sending/receiving so we don't need to do it here
	return 0;
}
